@@ -3,11 +3,13 @@
 {
     'name': 'Stock Supply Management',
     'name_bg_BG': 'Управление на доставка на наличност',
+    'name_ca_ES': 'Gestió del proveïment d''estocs',
     'name_de_DE': 'Lagerverwaltung Bestellwesen',
+    'name_es_AR': 'Gestión de suministro de existencias',
     'name_es_CO': 'Gestión de pedidos de inventario',
-    'name_es_ES': 'Gestión de suministro de existencias',
+    'name_es_ES': 'Gestión del abastecimiento de stock',
     'name_fr_FR': 'Gestion des approvisionnements de stock',
-    'version': '2.2.0',
+    'version': '2.4.0',
     'author': 'B2CK',
     'email': 'info@b2ck.com',
     'website': 'http://www.tryton.org/',
@@ -26,6 +28,14 @@ With schedulers:
 С планировщици:
     - за генериране на заявки за покупка въз основа на пренареждания
     - за генериране на вътрешни пратки въз основа на пренареждания
+''',
+    'description_ca_ES': '''Mòdul per la gestió del proveïment d'estocs amb:
+    - Estocs mínims
+    - Peticions de compra
+
+I planificadors:
+    - Per generar peticions de compra basades en estocs mínims
+    - Per generar albarans interns basats en estocs mínims
 ''',
     'description_de_DE': '''Modul für das Bestellwesen mit:
     - Bestellpunkten
@@ -34,6 +44,14 @@ With schedulers:
 Mit automatischer Auftragserstellung per Zeitplaner:
     - für Bestellungen auf der Basis von Bestellpunkten
     - für interne Lieferposten auf der Basis von Bestellpunkten
+''',
+    'description_es_AR': '''Módulo de gestión de suministros con:
+    - Punto de orden
+    - Solicitud de compra
+
+Con tareas programadas para:
+    - generar solicitudes de compra basados en puntos de orden
+    - generar envios internos basados en puntos de orden
 ''',
     'description_es_CO': '''Módulo de Administración de pedidos:
     - Punto de Orden
@@ -43,13 +61,13 @@ Con agendadores para:
     - generar solicitudes de compra basados en puntos de orden
     - generar empaques internos basados en puntos de orden
 ''',
-    'description_es_ES': '''Módulo de gestión de suministros con:
-    - Punto de orden
-    - Solicitud de compra
+    'description_es_ES': '''Módulo para la gestión de abastecimientos con:
+    - Reglas de stock mínimo
+    - Solicitudes de compra
 
-Con tareas programadas para:
-    - generar solicitudes de compra basados en puntos de orden
-    - generar envios internos basados en puntos de orden
+Con planificadores para:
+    - Generar solicitudes de compra basadas en stocks mínimos
+    - Generar envíos internos basados en stocks mínimos
 ''',
     'description_fr_FR': '''Module de gestion des approvisionnements avec:
     - Règles d'approvisionnement
@@ -77,7 +95,9 @@ sur base des règles d'approvisionnement
     'translation': [
         'locale/cs_CZ.po',
         'locale/bg_BG.po',
+        'locale/ca_ES.po',
         'locale/de_DE.po',
+        'locale/es_AR.po',
         'locale/es_CO.po',
         'locale/es_ES.po',
         'locale/fr_FR.po',
