@@ -2,8 +2,12 @@
 #this repository contains the full copyright notices and license terms.
 import datetime
 import operator
+from itertools import groupby
+from functools import partial
+from decimal import Decimal
 from trytond.model import ModelView, ModelSQL, fields
-from trytond.wizard import Wizard
+from trytond.wizard import Wizard, StateView, StateAction, StateTransition, \
+    Button
 from trytond.pyson import If, In, Eval, Get
 from trytond.transaction import Transaction
 from trytond.pool import Pool
@@ -14,22 +18,28 @@ class PurchaseRequest(ModelSQL, ModelView):
     _name = 'purchase.request'
     _description = __doc__
 
-    product = fields.Many2One(
-        'product.product', 'Product', required=True, select=1, readonly=True,
-        domain=[('purchasable', '=', True)])
-    party = fields.Many2One('party.party', 'Party',  select=1)
+    product = fields.Many2One('product.product', 'Product', required=True,
+        select=True, readonly=True, domain=[('purchasable', '=', True)])
+    party = fields.Many2One('party.party', 'Party',  select=True)
     quantity = fields.Float('Quantity', required=True)
-    uom = fields.Many2One('product.uom', 'UOM', required=True, select=1)
+    uom = fields.Many2One('product.uom', 'UOM', required=True, select=True)
     computed_quantity = fields.Float('Computed Quantity', readonly=True)
-    computed_uom = fields.Many2One('product.uom', 'Computed UOM', readonly=True)
+    computed_uom = fields.Many2One('product.uom', 'Computed UOM',
+        readonly=True)
     purchase_date = fields.Date('Best Purchase Date', readonly=True)
     supply_date = fields.Date('Expected Supply Date', readonly=True)
-    stock_level =  fields.Float('Stock at Supply Date', readonly=True)
+    stock_level = fields.Float('Stock at Supply Date', readonly=True)
     warehouse = fields.Many2One(
-        'stock.location', "Warehouse", required=True,
-        domain=[('type', '=', 'warehouse')], readonly=True)
-    purchase_line = fields.Many2One(
-        'purchase.line', 'Purchase Line',readonly=True)
+        'stock.location', "Warehouse",
+        states={
+            'required': Eval('warehouse_required', False),
+            },
+        domain=[('type', '=', 'warehouse')], depends=['warehouse_required'],
+        readonly=True)
+    warehouse_required = fields.Function(fields.Boolean('Warehouse Required'),
+        'get_warehouse_required')
+    purchase_line = fields.Many2One('purchase.line', 'Purchase Line',
+        readonly=True)
     purchase = fields.Function(fields.Many2One('purchase.purchase',
         'Purchase'), 'get_purchase')
     company = fields.Many2One('company.company', 'Company', required=True,
@@ -50,8 +60,13 @@ class PurchaseRequest(ModelSQL, ModelView):
         super(PurchaseRequest, self).__init__()
         self._order[0] = ('id', 'DESC')
         self._error_messages.update({
-            'create_request': 'Purchase requests are only created by the system.',
-            })
+                'create_request': 'Purchase requests are only created ' \
+                    'by the system.',
+                })
+        self._sql_constraints += [
+            ('check_purchase_request_quantity', 'CHECK(quantity > 0)',
+                'The requested quantity must be greater than 0'),
+            ]
 
     def init(self, module_name):
         cursor = Transaction().cursor
@@ -67,7 +82,10 @@ class PurchaseRequest(ModelSQL, ModelView):
             ids = [ids]
         res = {}
         for pr in self.browse(ids):
-            res[pr.id] = "%s@%s" % (pr.product.name, pr.warehouse.name)
+            if pr.warehouse:
+                res[pr.id] = "%s@%s" % (pr.product.name, pr.warehouse.name)
+            else:
+                res[pr.id] = pr.product.name
         return res
 
     def search_rec_name(self, name, clause):
@@ -79,7 +97,7 @@ class PurchaseRequest(ModelSQL, ModelView):
         return res
 
     def default_company(self):
-        return Transaction().context.get('company') or False
+        return Transaction().context.get('company')
 
     def get_purchase(self, ids, name):
         res = {}
@@ -89,7 +107,7 @@ class PurchaseRequest(ModelSQL, ModelView):
             if request.purchase_line:
                 res[request.id] = request.purchase_line.purchase.id
             else:
-                res[request.id] = False
+                res[request.id] = None
         return res
 
     def get_state(self, ids, name):
@@ -104,6 +122,12 @@ class PurchaseRequest(ModelSQL, ModelView):
                     res[request.id] = 'purchased'
         return res
 
+    def get_warehouse_required(self, ids, name):
+        requireds = {}
+        for request in self.browse(ids):
+            requireds[request.id] = request.product.type in ('goods', 'assets')
+        return requireds
+
     def origin_get(self):
         model_obj = Pool().get('ir.model')
         res = []
@@ -121,7 +145,6 @@ class PurchaseRequest(ModelSQL, ModelView):
         """
         pool = Pool()
         order_point_obj = pool.get('stock.order_point')
-        purchase_request_obj = pool.get('purchase.request')
         product_obj = pool.get('product.product')
         location_obj = pool.get('stock.location')
         user_obj = pool.get('res.user')
@@ -129,8 +152,8 @@ class PurchaseRequest(ModelSQL, ModelView):
 
         # fetch warehouses:
         warehouse_ids = location_obj.search([
-            ('type','=','warehouse'),
-            ])
+                ('type', '=', 'warehouse'),
+                ])
         # fetch order points
         order_point_ids = order_point_obj.search([
             ('type', '=', 'purchase'),
@@ -142,40 +165,48 @@ class PurchaseRequest(ModelSQL, ModelView):
                 (order_point.warehouse_location.id, order_point.product.id)
                 ] = order_point
 
-        # fetch stockable products
+        # fetch goods
         product_ids = product_obj.search([
-            ('type', '=', 'stockable'), 
-            ('purchasable', '=', True),
-            ])
+                ('type', '=', 'goods'),
+                ('consumable', '=', False),
+                ('purchasable', '=', True),
+                ])
         #aggregate product by minimum supply date
         date2products = {}
         for product in product_obj.browse(product_ids):
             min_date, max_date = self.get_supply_dates(product)
-            date2products.setdefault(min_date, []).append((product, max_date))
+            date2products.setdefault((min_date, max_date), []).append(product)
 
         # compute requests
         new_requests = []
-        for min_date in date2products:
-            product_ids = [x[0].id for x in date2products[min_date]]
-            with Transaction().set_context(forecast=True,
-                    stock_date_end=min_date or datetime.date.max):
-                pbl = product_obj.products_by_location(warehouse_ids,
-                    product_ids, with_childs=True, skip_zero=False)
-            for product, max_date in date2products[min_date]:
+        cursor = Transaction().cursor
+        for dates, products in date2products.iteritems():
+            min_date, max_date = dates
+            for i in range(0, len(products), cursor.IN_MAX):
+                product_ids = [p.id for p in products[i:i + cursor.IN_MAX]]
+                with Transaction().set_context(forecast=True,
+                        stock_date_end=min_date or datetime.date.max):
+                    pbl = product_obj.products_by_location(warehouse_ids,
+                        product_ids, with_childs=True, skip_zero=False)
                 for warehouse_id in warehouse_ids:
-                    qty = pbl.pop((warehouse_id, product.id))
-                    order_point = product2ops.get((warehouse_id, product.id))
+                    min_date_qties = dict((x, pbl.pop((warehouse_id, x)))
+                        for x in product_ids)
                     # Search for shortage between min-max
-                    shortage_date, product_quantity = self.get_shortage(
-                        warehouse_id, product.id, min_date, max_date,
-                        min_date_qty=qty, order_point=order_point)
+                    shortages = self.get_shortage(warehouse_id, product_ids,
+                        min_date, max_date, min_date_qties=min_date_qties,
+                        order_points=product2ops)
 
-                    if shortage_date == None or product_quantity == None:
-                        continue
-                    # generate request values
-                    request_val = self.compute_request(product, warehouse_id,
-                        shortage_date, product_quantity, company, order_point)
-                    new_requests.append(request_val)
+                    for product in products[i:i + cursor.IN_MAX]:
+                        shortage_date, product_quantity = shortages[product.id]
+                        if shortage_date == None or product_quantity == None:
+                            continue
+                        order_point = product2ops.get(
+                            (warehouse_id, product.id))
+                        # generate request values
+                        request_val = self.compute_request(product,
+                            warehouse_id, shortage_date, product_quantity,
+                            company, order_point)
+                        new_requests.append(request_val)
 
         new_requests = self.compare_requests(new_requests)
 
@@ -189,12 +220,13 @@ class PurchaseRequest(ModelSQL, ModelView):
             if new_req['supply_date'] == datetime.date.max:
                 new_req['supply_date'] = None
             if new_req['quantity'] > 0.0:
-                new_req.update({'product': new_req['product'].id,
-                                'party': new_req['party'] and new_req['party'].id,
-                                'uom': new_req['uom'].id,
-                                'computed_uom': new_req['computed_uom'].id,
-                                'company': new_req['company'].id
-                                })
+                new_req.update({
+                        'product': new_req['product'].id,
+                        'party': new_req['party'] and new_req['party'].id,
+                        'uom': new_req['uom'].id,
+                        'computed_uom': new_req['computed_uom'].id,
+                        'company': new_req['company'].id
+                        })
                 request_obj.create(new_req)
 
     def compare_requests(self, new_requests):
@@ -206,15 +238,14 @@ class PurchaseRequest(ModelSQL, ModelView):
         pool = Pool()
         uom_obj = pool.get('product.uom')
         request_obj = pool.get('purchase.request')
-        product_supplier_obj = pool.get('purchase.product_supplier')
         req_ids = request_obj.search([
-            ('purchase_line', '=', False),
+            ('purchase_line', '=', None),
             ('origin', 'like', 'stock.order_point,%'),
             ])
         request_obj.delete(req_ids)
 
         req_ids = request_obj.search([
-                ('purchase_line.moves', '=', False),
+                ('purchase_line.moves', '=', None),
                 ('purchase_line.purchase.state', '!=', 'cancel'),
                 ('origin', 'like', 'stock.order_point,%'),
                 ])
@@ -239,13 +270,15 @@ class PurchaseRequest(ModelSQL, ModelView):
 
             existing_req.setdefault(
                 (request.product.id, request.warehouse.id),
-                []).append({'supply_date': request.supply_date or datetime.date.max,
-                            'quantity': quantity,
-                            'uom': uom}
-                           )
+                []).append({
+                        'supply_date': \
+                            request.supply_date or datetime.date.max,
+                        'quantity': quantity,
+                        'uom': uom,
+                        })
 
         for i in existing_req.itervalues():
-            i.sort(lambda r,s: cmp(r['supply_date'],s['supply_date']))
+            i.sort(lambda r, s: cmp(r['supply_date'], s['supply_date']))
 
         # Update new requests to take existing requests into account
         new_requests.sort(key=operator.itemgetter('supply_date'))
@@ -255,7 +288,8 @@ class PurchaseRequest(ModelSQL, ModelView):
                 if old_req['supply_date'] <= new_req['supply_date']:
                     quantity = uom_obj.compute_qty(old_req['uom'],
                             old_req['quantity'], new_req['uom'])
-                    new_req['quantity'] = max(0.0, new_req['quantity'] - quantity)
+                    new_req['quantity'] = max(0.0,
+                        new_req['quantity'] - quantity)
                     new_req['computed_quantity'] = new_req['quantity']
                     old_req['quantity'] = uom_obj.compute_qty(new_req['uom'],
                         max(0.0, quantity - new_req['quantity']),
@@ -280,8 +314,13 @@ class PurchaseRequest(ModelSQL, ModelView):
         today = date_obj.today()
 
         for product_supplier in product.product_suppliers:
-            supply_date, next_supply_date = product_supplier_obj.\
-                    compute_supply_date(product_supplier, date=today)
+            supply_date = product_supplier_obj.compute_supply_date(
+                    product_supplier, date=today)
+            # TODO next_day is by default today + 1 but should depends
+            # on the CRON activity
+            next_day = today + datetime.timedelta(1)
+            next_supply_date = product_supplier_obj.compute_supply_date(
+                            product_supplier, date=next_day)
             if (not min_date) or supply_date < min_date:
                 min_date = supply_date
             if (not max_date):
@@ -310,7 +349,7 @@ class PurchaseRequest(ModelSQL, ModelView):
         today = date_obj.today()
         for product_supplier in product.product_suppliers:
             supply_date = product_supplier_obj.compute_supply_date(
-                    product_supplier, date=today)[0]
+                    product_supplier, date=today)
             sup_timedelta = date - supply_date
             if not supplier:
                 supplier = product_supplier.party
@@ -350,7 +389,7 @@ class PurchaseRequest(ModelSQL, ModelView):
                 product.purchase_uom or product.default_uom)
 
         if order_point:
-            origin = 'stock.order_point,%s'%order_point.id
+            origin = 'stock.order_point,%s' % order_point.id
         else:
             origin = 'stock.order_point,-1'
         return {'product': product,
@@ -367,54 +406,62 @@ class PurchaseRequest(ModelSQL, ModelView):
                 'origin': origin,
                 }
 
-    def get_shortage(self, location_id, product_id, min_date,
-                     max_date, min_date_qty, order_point=None):
+    def get_shortage(self, location_id, product_ids, min_date, max_date,
+            min_date_qties, order_points):
         """
-        Return between min_date and max_date  the first date where the
-            stock quantity is less than the minimal quantity and
-            the smallest stock quantity in the interval
-            or None if there is no date where stock quantity is less than
-            the minimal quantity
-        The minimal quantity comes from the order_point or is zero
+        Return for each product the first date between min_date and max_date
+        where the stock quantity is less than the minimal quantity and the
+        smallest stock quantity in the interval or None if there is no date
+        where stock quantity is less than the minimal quantity.
 
-        :param location_id: the stock location id
-        :param produc_id: the product id
-        :param min_date: the minimal date
-        :param max_date: the maximal date
-        :param min_date_qty: the stock quantity at the minimal date
-        :param order_point: a BrowseRecord of the Order Point
-        :return: a tuple with the date and the quantity
+        The minimal quantity comes from the order point or is zero.
+
+        min_date_qty is the quantities for each products at the min_date.
+        order_points is a dictionary that links products to order point.
         """
         product_obj = Pool().get('product.product')
 
-        res_date = None
-        res_qty = None
+        res_dates = {}
+        res_qties = {}
 
-        min_quantity = order_point and order_point.min_quantity or 0.0
+        min_quantities = {}
+        for product_id in product_ids:
+            order_point = order_points.get((location_id, product_id))
+            if order_point:
+                min_quantities[product_id] = order_point.min_quantity
+            else:
+                min_quantities[product_id] = 0.0
 
         current_date = min_date
-        current_qty = min_date_qty
+        current_qties = min_date_qties.copy()
         while (current_date < max_date) or (current_date == min_date):
-            if current_qty < min_quantity:
-                if not res_date:
-                    res_date = current_date
-                if (not res_qty) or (current_qty < res_qty):
-                    res_qty = current_qty
+            for product_id in product_ids:
+                current_qty = current_qties[product_id]
+                min_quantity = min_quantities[product_id]
+                res_qty = res_qties.get(product_id)
+                res_date = res_dates.get(product_id)
+                if current_qty < min_quantity:
+                    if not res_date:
+                        res_dates[product_id] = current_date
+                    if (not res_qty) or (current_qty < res_qty):
+                        res_qties[product_id] = current_qty
 
             with Transaction().set_context(stock_date_start=current_date,
                     stock_date_end=current_date):
-                res = product_obj.products_by_location([location_id],
-                [product_id], with_childs=True, skip_zero=False)
-            for qty in res.itervalues():
-                current_qty += qty
+                pbl = product_obj.products_by_location([location_id],
+                    product_ids, with_childs=True, skip_zero=False)
+            for key, qty in pbl.iteritems():
+                _, product_id = key
+                current_qties[product_id] += qty
             if current_date == datetime.date.max:
                 break
             current_date += datetime.timedelta(1)
 
-        return (res_date, res_qty)
+        return dict((x, (res_dates.get(x), res_qties.get(x)))
+            for x in product_ids)
 
     def create(self, vals):
-        for field_name in ('product', 'quantity', 'uom', 'warehouse', 'company'):
+        for field_name in ('product', 'quantity', 'uom', 'company'):
             if not vals.get(field_name):
                 self.raise_user_error('create_request')
         return super(PurchaseRequest, self).create(vals)
@@ -422,51 +469,32 @@ class PurchaseRequest(ModelSQL, ModelView):
 PurchaseRequest()
 
 
-class CreatePurchaseRequestInit(ModelView):
-    'Create Purchase Request Init'
-    _name = 'purchase.request.create_purchase_request.init'
+class CreatePurchaseRequestStart(ModelView):
+    'Create Purchase Request'
+    _name = 'purchase.request.create.start'
     _description = __doc__
 
-CreatePurchaseRequestInit()
+CreatePurchaseRequestStart()
 
 
 class CreatePurchaseRequest(Wizard):
     'Create Purchase Request'
-    _name = 'purchase.request.create_purchase_request'
+    _name = 'purchase.request.create'
 
-    states = {
-        'init': {
-            'result': {
-                'type': 'form',
-                'object': 'purchase.request.create_purchase_request.init',
-                'state': [
-                    ('end', 'Cancel', 'tryton-cancel'),
-                    ('create', 'Create', 'tryton-ok', True),
-                    ],
-                },
-            },
-        'create': {
-            'actions': ['_create_purchase_request'],
-            'result': {
-                'type': 'action',
-                'action': '_open',
-                'state': 'end',
-                },
-            },
-        }
+    start = StateView('purchase.request.create.start',
+        'stock_supply.purchase_request_create_start_view_form', [
+            Button('Cancel', 'end', 'tryton-cancel'),
+            Button('Create', 'create_', 'tryton-ok', default=True),
+            ])
+    create_ = StateAction('stock_supply.act_purchase_request_form_draft')
 
-    def _create_purchase_request(self, data):
+    def do_create_(self, session, action):
         purchase_request_obj = Pool().get('purchase.request')
         purchase_request_obj.generate_requests()
-        return {}
+        return action, {}
 
-    def _open(self, data):
-        pool = Pool()
-        model_data_obj = pool.get('ir.model.data')
-        act_window_obj = pool.get('ir.action.act_window')
-        act_window_id = model_data_obj.get_id('stock_supply',
-            'act_purchase_request_form_draft')
-        return act_window_obj.read(act_window_id)
+    def transition_create_(self, session):
+        return 'end'
 
 CreatePurchaseRequest()
 
@@ -482,6 +510,7 @@ class CreatePurchaseAskTerm(ModelView):
 
 CreatePurchaseAskTerm()
 
+
 class CreatePurchaseAskParty(ModelView):
     'Create Purchase Ask Party'
     _name = 'purchase.request.create_purchase.ask_party'
@@ -492,165 +521,163 @@ class CreatePurchaseAskParty(ModelView):
 
 CreatePurchaseAskParty()
 
+
 class CreatePurchase(Wizard):
     'Create Purchase'
     _name = 'purchase.request.create_purchase'
 
-    states = {
-
-        'init': {
-            'result': {
-                'type': 'choice',
-                'next_state': '_create_purchase',
-                },
-            },
-
-
-        'ask_user_party': {
-            'actions': ['_set_default_party'],
-            'result': {
-                'type': 'form',
-                'object': 'purchase.request.create_purchase.ask_party',
-                'state': [
-                    ('end', 'Cancel', 'tryton-cancel'),
-                    ('init', 'Continue', 'tryton-go-next', True),
-                    ],
-                },
-            },
-
-        'ask_user_term': {
-            'actions': ['_set_default_term'],
-            'result': {
-                'type': 'form',
-                'object': 'purchase.request.create_purchase.ask_term',
-                'state': [
-                    ('end', 'Cancel', 'tryton-cancel'),
-                    ('init', 'Continue', 'tryton-go-next', True),
-                    ],
-                },
-            },
-
-        }
+    start = StateTransition()
+    ask_party = StateView('purchase.request.create_purchase.ask_party',
+        'stock_supply.purchase_request_create_purchase_ask_party', [
+            Button('Cancel', 'end', 'tryton-cancel'),
+            Button('Continue', 'start', 'tryton-go-next', default=True),
+            ])
+    ask_term = StateView('purchase.request.create_purchase.ask_term',
+        'stock_supply.purchase_request_create_purchase_ask_term_view_form', [
+            Button('Cancel', 'end', 'tryton-cancel'),
+            Button('Continue', 'start', 'tryton-go-next', default=True),
+            ])
 
     def __init__(self):
         super(CreatePurchase, self).__init__()
         self._error_messages.update({
-            'missing_price': 'Purchase price is missing for product: %s (id: %s)!',
+            'missing_price': 'Purchase price is missing for product: %s ' \
+                '(id: %s)!',
             'please_update': 'This price is necessary for creating purchase.'
             })
 
-    def _set_default_party(self, data):
-
+    def default_ask_party(self, session, fields):
         request_obj = Pool().get('purchase.request')
-        requests = request_obj.browse(data['ids'])
+        requests = request_obj.browse(Transaction().context['active_ids'])
         for request in requests:
             if request.purchase_line:
                 continue
             if not request.party:
-                return {'product': request.product.id,'company': request.company.id}
-
-        return {'product': request.product.id,'company': request.company.id}
-
-    def _set_default_term(self, data):
+                return {
+                    'product': request.product.id,
+                    'company': request.company.id,
+                    }
+        return {
+            'product': request.product.id,
+            'company': request.company.id,
+            }
 
+    def default_ask_term(self, session, fields):
         request_obj = Pool().get('purchase.request')
-        requests = request_obj.browse(data['ids'])
+        requests = request_obj.browse(Transaction().context['active_ids'])
         for request in requests:
             if (not request.party) or request.purchase_line:
                 continue
             if not request.party.supplier_payment_term:
-                return {'party': request.party.id,'company': request.company.id}
+                return {
+                    'party': request.party.id,
+                    'company': request.company.id,
+                    }
+        return {
+            'party': request.party.id,
+            'company': request.company.id,
+            }
 
-        return {'party': request.party.id,'company': request.company.id}
+    def _group_purchase_key(self, requests, request):
+        '''
+        The key to group lines by purchases
 
-    def _create_purchase(self, data):
+        :param requests: a list of requests
+        :param request: the request
+
+        :return: a list of key-value as tuples of the purchase
+        '''
+        party_obj = Pool().get('party.party')
+
+        return (
+            ('company', request.company.id),
+            ('party', request.party.id),
+            ('payment_term', request.party.supplier_payment_term.id),
+            ('warehouse', request.warehouse.id),
+            # XXX use function field
+            ('currency', request.company.currency.id),
+            ('invoice_address', party_obj.address_get(request.party.id,
+                    type='invoice')),
+            )
+
+    def transition_start(self, session):
         pool = Pool()
         request_obj = pool.get('purchase.request')
         party_obj = pool.get('party.party')
         purchase_obj = pool.get('purchase.purchase')
-        product_obj = pool.get('product.product')
         line_obj = pool.get('purchase.line')
         date_obj = pool.get('ir.date')
 
-        form = data['form']
-        if form.get('product') and form.get('party') and \
-                form.get('company'):
-            req_ids = request_obj.search([
-                ('id', 'in', data['ids']),
-                ('party', '=', False),
-                ])
-            if req_ids:
-                request_obj.write(req_ids, {'party': form['party']})
+        request_ids = Transaction().context['active_ids']
 
-        elif form.get('payment_term') and form.get('party') and \
-                form.get('company'):
-            with Transaction().set_context(company=form['company']):
-                party_obj.write(form['party'],{
-                    'supplier_payment_term': form['payment_term']
-                    })
+        if (session.ask_party.product
+                and session.ask_party.party
+                and session.ask_party.company):
+            req_ids = request_obj.search([
+                    ('id', 'in', request_ids),
+                    ('party', '=', None),
+                    ])
+            if req_ids:
+                request_obj.write(req_ids, {
+                        'party': session.ask_party.party.id,
+                        })
+            session.ask_party.product = None
+            session.ask_party.party = None
+            session.ask_party.company = None
+        elif (session.ask_term.payment_term
+                and session.ask_term.party
+                and session.ask_term.company):
+            with Transaction().set_context(
+                    company=session.ask_term.company.id):
+                party_obj.write(session.ask_term.party.id, {
+                        'supplier_payment_term': \
+                            session.ask_term.payment_term.id,
+                        })
+            session.ask_term.payment_term = None
+            session.ask_term.party = None
+            session.ask_term.company = None
 
         req_ids = request_obj.search([
-            ('id', 'in', data['ids']),
-            ('purchase_line', '=', False),
-            ('party', '=', False),
-            ])
+                ('id', 'in', request_ids),
+                ('purchase_line', '=', None),
+                ('party', '=', None),
+                ])
         if req_ids:
-            return 'ask_user_party'
+            return 'ask_party'
 
         today = date_obj.today()
-        requests = request_obj.browse(data['ids'])
-        purchases = {}
-        # collect data
-        for request in requests:
-            if request.purchase_line:
-                continue
+        requests = request_obj.browse(request_ids)
 
-            if not request.party.supplier_payment_term:
-                return 'ask_user_term'
+        requests = [r for r in requests if not r.purchase_line]
+        if any(r for r in requests if not r.party.supplier_payment_term):
+            return 'ask_term'
 
-            key = (request.party.id, request.company.id, request.warehouse.id)
-            if key not in purchases:
-                if request.purchase_date and request.purchase_date >= today:
-                    purchase_date = request.purchase_date
-                else:
+        keyfunc = partial(self._group_purchase_key, requests)
+        requests = sorted(requests, key=keyfunc)
+
+        with Transaction().set_user(0, set_context=True):
+            for key, grouped_requests in groupby(requests, key=keyfunc):
+                grouped_requests = list(grouped_requests)
+                try:
+                    purchase_date = min(r.purchase_date
+                        for r in grouped_requests
+                        if r.purchase_date)
+                except ValueError:
                     purchase_date = today
-                purchase = {
-                    'company': request.company.id,
-                    'party': request.party.id,
+                if purchase_date < today:
+                    purchase_date = today
+                values = {
                     'purchase_date': purchase_date,
-                    'payment_term': request.party.supplier_payment_term.id,
-                    'warehouse': request.warehouse.id,
-                    'currency': request.company.currency.id,
-                    'invoice_address': party_obj.address_get(request.party.id,
-                            type='invoice'),
-                    'lines': [],
                     }
-
-                purchases[key] = purchase
-            else:
-                purchase = purchases[key]
-
-            line = self.compute_purchase_line(request)
-            purchase['lines'].append(line)
-            if request.purchase_date:
-                if purchase.get('purchase_date'):
-                    purchase['purchase_date'] = min(purchase['purchase_date'],
-                                                    request.purchase_date)
-                else:
-                    purchase['purchase_date'] = request.purchase_date
-
-        # Create all
-        for purchase in purchases.itervalues():
-            lines = purchase.pop('lines')
-            with Transaction().set_user(0, set_context=True):
-                purchase_id = purchase_obj.create(purchase)
-            for line in lines:
-                request_id = line.pop('request')
-                line['purchase'] = purchase_id
-                with Transaction().set_user(0):
-                    line_id = line_obj.create(line)
-                request_obj.write(request_id, {'purchase_line': line_id})
-
+                values.update(dict(key))
+                purchase_id = purchase_obj.create(values)
+                for request in grouped_requests:
+                    values = self.compute_purchase_line(request)
+                    values['purchase'] = purchase_id
+                    line_id = line_obj.create(values)
+                    request_obj.write(request.id, {
+                            'purchase_line': line_id,
+                            })
         return 'end'
 
     def _get_tax_rule_pattern(self, request):
@@ -665,15 +692,14 @@ class CreatePurchase(Wizard):
 
     def compute_purchase_line(self, request):
         pool = Pool()
-        party_obj = pool.get('party.party')
         product_obj = pool.get('product.product')
         tax_rule_obj = pool.get('account.tax.rule')
+        line_obj = pool.get('purchase.line')
 
         line = {
             'product': request.product.id,
             'unit': request.uom.id,
             'quantity': request.quantity,
-            'request': request.id,
             'description': request.product.name,
             }
 
@@ -683,6 +709,8 @@ class CreatePurchase(Wizard):
                 currency=request.company.currency.id):
             product_price = product_obj.get_purchase_price(
                     [request.product.id], request.quantity)[request.product.id]
+            product_price = product_price.quantize(
+                Decimal(1) / 10 ** line_obj.unit_price.digits[1])
 
         if not product_price:
             self.raise_user_error('missing_price', (request.product.name,
@@ -693,7 +721,7 @@ class CreatePurchase(Wizard):
         for tax in request.product.supplier_taxes_used:
             if request.party and request.party.supplier_tax_rule:
                 pattern = self._get_tax_rule_pattern(request)
-                tax_id = tax_rule_obj.apply(request.party.supplier_tax_rule, 
+                tax_id = tax_rule_obj.apply(request.party.supplier_tax_rule,
                         tax, pattern)
                 if tax_id:
                     taxes.append(tax_id)
